#
# This file is Copyright by the GPSD project
# SPDX-License-Identifier: BSD-2-Clause
#
"""Module to interface with gpsd."""

import gps

if __name__ == "__main__":
    """Convenient way to get module version"""
    print("gps module version: %s" % gps.__version__)
